#!/usr/bin/perl

use v5.012;
use strict;
use warnings;

use Cwd;
use Test::More;

my $prog = $ENV{XMLMON} // './xmlmon';
my %files;

for my $flavor (qw(nx bad missing too-much good single-line cyrillic)) {
	state $test_dir = $ENV{TEST_DIR} // getcwd.'/t';
	my $up = uc $flavor;
	my $basename = $ENV{"TEST_${up}_NAME"} // "$flavor.xml";
	$files{$flavor} = $ENV{"TEST_${up}"} // "file://$test_dir/$basename";
}

my @patterns = qw( -p /main/real/thing -p /main/another/thing -p /main/някакво/нещо );

my @test_data = (
	{
		name => 'version and usage',
		args => [ qw(-h -V) ],
		lines => -1,
		first => qr/^ xmlmon \s /x,
	},
	{
		name => 'no patterns',
		args => [ $files{good} ],
		err => 'No pattern',
	},
	{
		name => 'pattern not starting with a slash',
		args => [ qw( -p foo ) ],
		err => 'Only patterns',
	},
	{
		name => 'invalid pattern - just a slash',
		args => [ qw( -p / ) ],
		err => 'Invalid pattern',
	},
	{
		name => 'invalid pattern - two slashes',
		args => [ qw( -p /foo//bar ) ],
		err => 'Invalid pattern',
	},
	{
		name => 'invalid pattern - trailing slash',
		args => [ qw( -p /foo/ )],
		err => 'Invalid pattern',
	},
	{
		name => 'no arguments',
		args => [ @patterns ],
		err => 'Exactly one',
	},
	{
		name => 'too many arguments',
		args => [ @patterns, $files{good}, $files{good} ],
		err => 'Exactly one',
	},
	{
		name => 'way too many arguments',
		args => [ @patterns, $files{good}, $files{good}, $files{good} ],
		err => 'Exactly one',
	},
	{
		name => 'nonexistent file',
		args => [ @patterns, $files{nx} ],
		err => 'Could not fetch',
	},
	{
		name => 'badly-formed XML',
		args => [ @patterns, $files{bad} ],
		err => 'Malformed',
	},
	{
		name => 'missing data',
		args => [ @patterns, $files{missing} ],
		err => 'No patterns found',
	},
	{
		name => 'unexpected elements inside',
		args => [ @patterns, $files{"too-much"} ],
		err => 'Unexpected',
	},
	{
		name => 'well-formed input',
		args => [ @patterns, $files{good} ],
		ok => '2',
	},
	{
		name => 'well-formed input on a single line',
		args => [ @patterns, $files{"single-line"} ],
		ok => '2',
	},
	{
		name => 'well-formed input with Cyrillic tags and data',
		args => [ @patterns, $files{cyrillic} ],
		ok => 'данни',
	},
);

sub check_wait_result($ $ $)
{
	my ($stat, $pid, $name) = @_;
	my $sig = $stat & 127;
	if ($sig != 0) {
		die "Program '$name' (pid $pid) was killed by signal $sig\n";
	} else {
		my $code = $stat >> 8;
		if ($code != 0) {
			die "Program '$name' (pid $pid) exited with non-zero status $code\n";
		}
	}
}

sub run_command(@)
{
	my (@cmd) = @_;
	my $name = $cmd[0];

	my $pid = open my $f, '-|';
	if (!defined $pid) {
		die "Could not fork for $name: $!\n";
	} elsif ($pid == 0) {
		exec { $name } @cmd;
		die "Could not execute '$name': $!\n";
	}
	my @res = <$f>;
	close $f;
	check_wait_result $?, $pid, $name;
	return @res;
}

sub test_xmlmon($)
{
	my ($data) = @_;
	my $test = "$data->{name}";

	my @lines = eval { run_command $prog, @{$data->{args}} };
	my $msg = $@;
	if ($data->{fail}) {
		isnt $msg, '', "$test: fails";
	} else {
		is $msg, '', "$test: succeeds";
	}
	SKIP:
	{
		skip "$test: failed, not matching the output", 2 if $msg ne '';

		SKIP:
		{
			my $expected = $data->{lines} // 1;
			skip "$test: indeterminate number of lines expected", 1 if $expected < 0;
			my $linestr = $expected == 1? 'line': 'lines';
			is scalar @lines, $expected, "$test: $expected $linestr of output";
		}

		if ($data->{first}) {
			like $lines[0], $data->{first},
			    "$test: proper first line";
		} elsif ($data->{err}) {
			like $lines[0], qr/^ ERR \s+ \Q$data->{err}\E /x,
			    "$test: error '$data->{err}'";
		} elsif ($data->{ok}) {
			like $lines[0], qr/^ OK \s+ \Q$data->{ok}\E \s* $ /x,
			    "$test: exact output";
		} else {
			like $lines[0], qr/^ OK \s /x,
			    "$test: okay";
		}
	}
}

plan tests => 3 * scalar @test_data;

test_xmlmon $_ for @test_data;
